"""Testing code generated by nbdev in unpackai/nlp/data.py"""
# Generated automatically from notebook nbs/41_nlp_data.ipynb

from unpackai.nlp.data import *

# Test Cell
# For Test Cases (might have duplicate import because it will be in a dedicated file)
from pathlib import Path
from typing import List

import pytest
from test_common.utils_4_tests import DATA_DIR
from test_utils import GITHUB_TEST_DATA_URL, check_connection_github

# Test Cell
LOCAL_TEST_TXT = DATA_DIR / "to_download.txt"
GITHUB_TEST_TXT = f"{GITHUB_TEST_DATA_URL}/to_download.txt"

LOCAL_TEST_TXT_UTF8 = DATA_DIR / "Deep learning.txt"
GITHUB_TEST_TXT_UTF8 = f"{GITHUB_TEST_DATA_URL}/Deep%20learning.txt"


@pytest.fixture(scope="session")
def test_txt_content():
    return LOCAL_TEST_TXT.read_text()


@pytest.fixture(scope="session")
def local_textual(test_txt_content):
    return Textual(test_txt_content)


class Test_Textual:
    def test_init(self, local_textual, test_txt_content):
        """Test initialization of Textual from text"""
        expected_txt = test_txt_content.replace("\n", " ").replace("\r", "")
        assert local_textual.text == expected_txt

    def test_init_encoding(self):
        """Test initialization of Textual from text"""
        content = LOCAL_TEST_TXT_UTF8.read_text(encoding="utf-8")
        textual = Textual(content)
        assert textual.text == content.replace("\n", " ").replace("\r", "")

    def test_from_path(self, local_textual):
        """Test create Textual from path (existing)"""
        textual = Textual.from_path(LOCAL_TEST_TXT)
        assert textual.text == local_textual.text

    def test_from_path_error(self):
        """Test extract Textual of file that does not exist"""
        with pytest.raises(FileExistsError):
            textual = Textual.from_path("does_not_exist.txt")

    def test_from_url(self, check_connection_github, local_textual):
        """Test extract Textual from URL"""
        textual = Textual.from_url(GITHUB_TEST_TXT)
        assert textual.text == local_textual.text, f"URL text: {textual.text}"

    def test_from_url_non_ascii(self, check_connection_github):
        """Test extract Textual from URL with non-ascii characters"""
        textual = Textual.from_url(GITHUB_TEST_TXT_UTF8)
        content = LOCAL_TEST_TXT_UTF8.read_text(encoding="utf-8")
        content = content.replace("\n", " ").replace("\r", "")
        assert textual.text == content, f"URL text: {textual.text}"



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/18_interp.ipynb (unless otherwise specified).

__all__ = []

# Cell
from ..utils import hush
from ..bug import ishell, BUGBOOK
from .latent import CosineSearch, InterpEmbeddings
from .vision import (
    image_upload,
    upload_and_visualize_cnn,
    IMAGENET_CLASSES,
    IMAGENET_MEAN,
    IMAGENET_STD)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/45_nlp_interp.ipynb (unless otherwise specified).

__all__ = ['InterpEmbeddingsTokenizer']

# Cell
from ..interp.latent import InterpEmbeddings
import pandas as pd

# Cell
class InterpEmbeddingsTokenizer(InterpEmbeddings):
    def __init__(self,
                 embedding_matrix,
                 tokenizer):
        """
        embedding_matrix: np.ndarray, embedding matrix of shape:
            (num_items, hidden_size)
        tokenizer: a huggingface tokenizer
        """
        super().__init__(
            embedding_matrix,
            dict((v, k) for k, v in tokenizer.vocab.items()))
        self.tokenizer = tokenizer

    def search(
        self,
        word: str,
        filter_special_token: bool = True,
        top_k: int = 20,
    ) -> pd.DataFrame:
        """
        search for similar words with embedding and
            tokenizer's encode/ decode
        """
        token_ids = self.tokenizer.encode(word)
        if filter_special_token:
            token_ids = list(t for t in token_ids if t > 110)

        # combine multiple tokens into 1
        vec = self.base[token_ids].mean(0)

        # distance search
        closest, similarity = self.cosine.search(vec, return_similarity=True)
        tokens = self.tokenizer.convert_ids_to_tokens(closest)
        return pd.DataFrame({
            "tokens": tokens,
            "idx": closest,
            "similarity": similarity}).head(top_k)


from pathlib import Path
import sys

def test_import():
    """Test import of unpackai"""
    unpackai_root = Path(__file__).parent.parent.parent
    sys.path.append(str(unpackai_root))
    import unpackai



# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"PathStr": "51_pdf_utils.ipynb",
         "PathURL": "10_utils.ipynb",
         "logger": "10_utils.ipynb",
         "check_img": "10_utils.ipynb",
         "clean_error_img": "10_utils.ipynb",
         "hush": "10_utils.ipynb",
         "static_root": "10_utils.ipynb",
         "STATIC": "10_utils.ipynb",
         "friendly_size": "10_utils.ipynb",
         "ls": "10_utils.ipynb",
         "gv": "10_utils.ipynb",
         "get_url_size": "10_utils.ipynb",
         "download": "10_utils.ipynb",
         "url_2_text": "10_utils.ipynb",
         "download_and_unpack": "10_utils.ipynb",
         "unpack_archive": "10_utils.ipynb",
         "read_csv_from_zip": "10_utils.ipynb",
         "BugBook": "12_bug.ipynb",
         "module_not_found_message1": "12_bug.ipynb",
         "module_not_found_message2": "12_bug.ipynb",
         "module_not_found_error_filter": "12_bug.ipynb",
         "BUGBOOK": "12_bug.ipynb",
         "render_download_button": "12_bug.ipynb",
         "custom_exc": "12_bug.ipynb",
         "itb": "12_bug.ipynb",
         "CosineSearch": "15_interp_latent.ipynb",
         "InterpEmbeddings": "15_interp_latent.ipynb",
         "IMAGENET_CLASSES": "16_interp_cnn.ipynb",
         "IMAGENET_MEAN": "16_interp_cnn.ipynb",
         "IMAGENET_STD": "16_interp_cnn.ipynb",
         "simple_to_tensor": "16_interp_cnn.ipynb",
         "basic_trans": "16_interp_cnn.ipynb",
         "tensor_to_8b": "16_interp_cnn.ipynb",
         "get_features_resnet": "16_interp_cnn.ipynb",
         "plot_contrary_overlap": "16_interp_cnn.ipynb",
         "visualize_conv_backbone": "16_interp_cnn.ipynb",
         "image_upload": "16_interp_cnn.ipynb",
         "upload_and_visualize_cnn": "16_interp_cnn.ipynb",
         "ImageLabeler": "20_Image_Labeler.ipynb",
         "SingleClassImageLabeler": "20_Image_Labeler.ipynb",
         "MultiClassImageLabeler": "20_Image_Labeler.ipynb",
         "no_missing_values": "35_tabular_data.ipynb",
         "plot_hist": "35_tabular_data.ipynb",
         "plot_feat_correlations": "35_tabular_data.ipynb",
         "plot_feat_dendrogram": "35_tabular_data.ipynb",
         "keyerror_in_tabular": "39_tabular.ipynb",
         "Textual": "41_nlp_data.ipynb",
         "HFTextBlock": "41_nlp_data.ipynb",
         "InterpEmbeddingsTokenizer": "45_nlp_interp.ipynb",
         "TextualPDF": "51_pdf_utils.ipynb"}

modules = ["utils.py",
           "bug.py",
           "interp/latent.py",
           "interp/vision.py",
           "interp/__init__.py",
           "turk/image.py",
           "tabular/data.py",
           "tabular/__init__.py",
           "nlp/data.py",
           "nlp/interp.py",
           "nlp/__init__.py",
           "pdf.py"]

doc_url = "https://unpackai.github.io/unpackai/"

git_url = "https://github.com/unpackai/unpackai/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/10_utils.ipynb (unless otherwise specified).

__all__ = ['PathStr', 'PathURL', 'check_img', 'clean_error_img', 'hush', 'static_root', 'STATIC', 'friendly_size', 'ls',
           'gv', 'get_url_size', 'download', 'url_2_text', 'download_and_unpack', 'unpack_archive', 'read_csv_from_zip']

# Cell
import logging
import requests
import shutil
import tarfile
import tempfile
from datetime import datetime
from mimetypes import types_map
from pathlib import Path
from typing import Callable, List, Union
from zipfile import ZipFile

import graphviz
import pandas as pd
from bs4 import BeautifulSoup
from IPython.core import magic
from IPython.display import Javascript
from ipywidgets import Output, VBox, Button, HTML
from PIL import Image
from tqdm.notebook import tqdm
from uuid import uuid4


PathStr = Union[Path, str]
PathURL = Union[Path, str]


# Internal Cell
logger = logging.getLogger("unpackai.utils")

# Cell
def check_img(
    img: Path,
    formats: List[str] = [".jpg", ".jpeg", ".png", ".bmp"],
) -> None:
    """
    Check on a single image,
    If it's quality is troublesome
        we unlink/ditch the image
    """
    img = Path(img)
    # check if this path is an image
    if img.suffix.lower().split("?")[0] not in formats:
        return

    try:
        # try to open that image and then (to check truncated image)
        # We need to include in a with block to close the image before deleting
        with Image.open(img) as im:
            im.load()
    except Exception:
        if img.exists():
            img.unlink()
            logging.warning(f"Removed erroneous img: {img}")
            return


def clean_error_img(
    path: Path,
    progress: bool = True,
) -> None:
    """
    - path: an image directory
    - progress: do we print out progress bar or not
        default True
    """
    path = Path(path)

    # check directory existence
    if path.exists() == False:
        raise FileExistsError(
            f"""path does not exists on:{path},
    make sure there is a directory "{path.name}".
    under directory "{path.parent}"
    """)

    # create iterator, probably with progress bar
    iterator = tqdm(list(path.iterdir()), leave=False)\
        if progress else path.iterdir()

    for obj in iterator:
        if obj.is_dir():
            # use recursion to clean the sub folder
            clean_error_img(obj, progress=progress)
        else:
            # cheking on a single image
            check_img(obj)

# Cell

@magic.register_cell_magic
def hush(line, cell):
    """
    A magic cell function to collapse the print out
    %%hush
    how_loud = 100
    how_verbose = "very"
    do_some_python_thing(how_loud, how_verbose)
    """
    # the current output widget
    out = Output()
    output_box = VBox([out])

    # create uuid for DOM identifying
    uuid = str(uuid4())

    # default setting is to hide the print out
    output_box.layout.display = "none"
    # the toggling button
    show_btn = Button(description="toggle output")
    show_btn.add_class(f"hush_toggle_{uuid}")
    output_box.add_class(f"hush_output_{uuid}")


    # A control panel containing a button
    # and the output box
    total_control = VBox([show_btn, output_box])
    display(total_control)

    # assign JS listener
    display(Javascript(f"""
    console.info("loading toggle event: {uuid}")
    const toggle_hush = (e) =>\u007b
        var op = document.querySelector(".hush_output_{uuid}");
        if(op.style.display=="none")\u007b
             op.style.display="block"

        \u007d else \u007b
            op.style.display="none"
        \u007d

    \u007d
    document.querySelector('.hush_toggle_{uuid}').onclick=toggle_hush
    """))

    with out:
        ishell = get_ipython()
        # excute the code in cell
        result = ishell.run_cell(
            cell, silent=False)

    # we still want the error to be proclaimed loudly
    if result.error_in_exec:
        logging.error(f"'{result.error_in_exec}' happened, breaking silence now")
        result.raise_error()

# Cell
def static_root() -> Path:
    """
    Return static path
    """
    import unpackai

    unpackai_path = Path(unpackai.__path__[0])
    if not unpackai_path.exists():
        egg_path = Path(f"{unpackai_path}.egg-link")
        unpackai_path = Path(egg_path.read_text())

    return unpackai_path / "static"


STATIC = static_root()

# Internal Cell
def _iter_files(root: Path, exclude_dir: List[str], hide_info: bool):
    """Return all info of files found in a root directory"""
    for f in root.rglob("*"):
        if any(d in f.parts for d in exclude_dir):
            continue

        info = {
            "Name": f.name,
            "Parent": f.parent.name,
            "Path": f.as_posix(),
            "Level": len(f.relative_to(root).parent.parts),
            "Last_Modif": datetime.fromtimestamp(f.stat().st_mtime),
        }
        if f.is_file():
            info.update(
                {
                    "FileDir": "File",
                    "Extension": f.suffix or f.name,
                    "Type": types_map.get(f.suffix.lower()),
                }
            )
            if not hide_info:
                size = f.stat().st_size
                info.update(
                    {
                        "Size": size,
                        "Friendly_Size": friendly_size(size),
                    }
                )
        elif f.is_dir():
            info.update({"FileDir": "Dir"})

        yield info


# Cell
def friendly_size(size: float) -> str:
    """Convert a size in bytes (as float) to a size with unit (as a string)"""
    unit = "B"
    # Reminder: 1 KB = 1024 B, and 1 MB = 1024 KB, ...
    for letter in "KMG":
        if size >= 1024:
            size /= 1024
            unit = f"{letter}B"

    # We want to keep 2 digits after floating point
    # because it is a good balance between precision and concision
    return f"{size:0.2f} {unit}"


def ls(root: PathStr, exclude: List[str] = None, hide_info=False) -> pd.DataFrame:
    """Return a DataFrame with list of files & directories in a given path.

    Args:
        root: root path to look for files and directories recursively
        exclude: optional list of names to exclude in the search (e.g. `[".git"]`)
    """
    df = pd.DataFrame(
        _iter_files(Path(root), exclude_dir=exclude or [], hide_info=hide_info)
    )
    return df.sort_values(by=["Path"])


# Cell
def gv(source: str, rankdir="LR"):
    """Generate a GraphViz diagram, with a settable orientation (i.e. rankdir)"""
    source = source.strip().rstrip(";")
    return graphviz.Source(f'digraph G {{  rankdir="{rankdir}" {source}; }}')


# Cell
def get_url_size(url: str) -> int:
    """Returns the size of URL, or -1 if it cannot get it"""
    with requests.request("HEAD", url) as resp:
        if resp.status_code != 200:
            raise ConnectionError(f"Error when retrieving size from {url}")

        return int(resp.headers.get("Content-Length", -1))


# Cell
def download(url: str, dest: PathStr = None) -> Path:
    """Download file and return the Path of the downloaded file

    If the destination is not specified, download in the current directory
    with the name specified in the URL
    """
    # We can allow empty destination, in which case we use file name in URL
    # We need to ensure that destination is a Path
    dest = Path(dest or url.rpartition("/")[-1])

    # For big files, we will split into chunks
    # We might also consider adding a progress bar
    size = get_url_size(url)
    if size < 1024 * 1024:
        with requests.get(url) as resp:
            resp.raise_for_status()
            dest.write_bytes(resp.content)

    else:
        with requests.get(url, stream=True) as resp:
            resp.raise_for_status()
            with dest.open("wb") as f:
                for chunk in resp.iter_content(chunk_size=8192):
                    f.write(chunk)

    logger.info(f"Downloaded {url} to {dest}")
    return dest


def url_2_text(url: str) -> str:
    """Extract text content from an URL (textual content for an HTML)"""
    resp = requests.get(url)
    if resp.status_code != 200:
        raise ConnectionError(f"Error when retrieving text content from {url}")

    resp.encoding = "utf-8"
    content_type = resp.headers["Content-Type"]
    if "html" in content_type:
        text = BeautifulSoup(resp.text).text
    else:
        text = resp.text
    return text


# Cell

# We will re-use directly the unzip function from shutil
# Since we want it in our "all", we just assign to itself so nbdev can detect it
unpack_archive = shutil.unpack_archive

# Equivalent to:
# def unpack_archive(filename, extract_dir=None, format=None):
#     return shutil.unpack_archive(filename, extract_dir=None, format=None)


def download_and_unpack(url: str, extract_dir: PathStr = None, fmt: str = None) -> Path:
    """Download a file and unzip it. Returns the path of unzipped directory

    Args:
        url: URL of the archive to download
        extract_dir: name of the target directory, where the archive is unpacked.
                    If not provided, <working directory> / <archive name without extesion> is used.
        fmt: archive format (one of "zip", "tar", "gztar", "bztar",or "xztar").
                Or any other registered format. If not provided, it is guessed from extension.
    """
    # We will store the downloaded archive in a temporary folder
    with tempfile.TemporaryDirectory() as tmpdirname:
        dest = Path(tmpdirname) / url.split("?")[0].rpartition("/")[-1]
        download(url, dest=dest)

        if extract_dir is None:
            extract_dir = Path.cwd() / dest.stem

        unpack_archive(str(dest), extract_dir=str(extract_dir), format=fmt)
        logger.info(f"Extracted {url} to folder {dest}")

    return Path(extract_dir)


# Internal Cell
def _zip_csv_2_df(zip_path:Path, csv_path:PathStr) -> pd.DataFrame:
    """CSV in zip to DataFrame"""
    with ZipFile(zip_path) as zf:
        try:
            with zf.open(str(csv_path)) as f_csv:
                return pd.read_csv(f_csv)
        except KeyError:
            files = "\n".join(f" * {f}" for f in zf.namelist() if f.lower().endswith(".csv"))
            raise FileNotFoundError(
                f'CSV file "{csv_path}" not found in "{zip_path}" '
                f"containing following CSV files:\n{files}"
            ) from None


def _tar_csv_2_df(tar_path:Path, csv_path:PathStr) -> pd.DataFrame:
    """CSV in tar to DataFrame"""
    with tarfile.open(tar_path) as tf:
        try:
            csv_member = tf.getmember(str(csv_path))
            return pd.read_csv(tf.extractfile(member=csv_member))
        except KeyError:
            files = "\n".join(f" * {f}" for f in tf.getnames() if f.lower().endswith(".csv"))
            raise FileNotFoundError(
                f'CSV file "{csv_path}" not found in "{tar_path}" '
                f"containing following CSV files:\n{files}"
            ) from None


# Cell
def read_csv_from_zip(archive: PathURL, csv_path: PathStr) -> pd.DataFrame:
    """Load a CSV file inside a zip/tar and returns the equivalent pandas `DataFrame`

    Args:
        archive: path or URL of the zip (or tar/tag.gz) file
        csv_path: path of csv relative to root of archive
            Note: if root is a folder, csv_path shall include this path (e.g. "archive/my_table.csv")
    """
    if Path(csv_path).suffix.lower() != ".csv":
        raise AttributeError(
            f'2nd argument ("csv_path") shall be a CSV file but is "{csv_path}"'
        )

    with tempfile.TemporaryDirectory() as tmpdirname:
        if isinstance(archive, str) and archive.startswith("http"):
            zip_path = Path(tmpdirname) / archive.split("?")[0].rpartition("/")[-1]
            download(archive, dest=zip_path)
        else:
            zip_path = Path(archive)

        extensions = "".join(zip_path.suffixes[-2:]).lower()
        if extensions == ".zip":
            return _zip_csv_2_df(zip_path, csv_path)
        elif extensions in (".tar", ".tar.gz"):
            return _tar_csv_2_df(zip_path, csv_path)
        else:
            raise AttributeError(
                f'Archive shall be either .zip, .tar, or .tar.gz but is "{zip_path}"'
            )

# Cell
try:
    ishell = get_ipython()
except NameError as e:
    from IPython.testing.globalipapp import get_ipython, start_ipython
    ishell = start_ipython()
    print(get_ipython())


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/39_tabular.ipynb (unless otherwise specified).

__all__ = ['keyerror_in_tabular']

# Cell
from ..utils import *
from ..bug import *
from .data import *

# Cell
from ..bug import ishell,BUGBOOK, itb

def keyerror_in_tabular(etype, evalue, tb):
    stb = itb.structured_traceback(etype, evalue, tb)
    sstb = itb.stb2text(stb)
    if "pandas/" in sstb:
        return f"We don't have column '{evalue}' in the dataframe"
    else:
        return None


BUGBOOK["KeyError"] = keyerror_in_tabular


"""Testing code generated by nbdev in unpackai/tabular/data.py"""
# Generated automatically from notebook nbs/35_tabular_data.ipynb

from unpackai.tabular.data import *

# Test Cell
import pandas as pd
from pathlib import Path
import pytest
import sys

try:
    root_dir = Path(".")
    df = pd.read_csv(root_dir / "test" / "tabular" / "california_housing.csv")
except FileNotFoundError as e:
    root_dir = Path("../")
    df = pd.read_csv(root_dir / "test" / "tabular" / "california_housing.csv")

# Test Cell
def test_no_missing_value_basic():
    from copy import deepcopy

    df2 = deepcopy(df).head(100)

    # create missing value manually
    for i, key in enumerate(df2):
        df2[key] = list(range(10 * (i + 1))) + [None] * (9 - i) * 10
    new_df = no_missing_values(df2)
    assert (
        list(new_df.columns) == list(df2.columns)[3:]
    ), f"""
        Incorrect column filtering
        based on missing value: {list(new_df.columns)}"""


# Test Cell
@pytest.mark.skipif(sys.platform != "linux", reason="plot test only for linux")
def test_plot_hist_run_through():
    """
    can plot_hist outpout an image
    """
    plot_hist(df.head(100))


# Test Cell
# hide
def repeat_cols(df, n=2):
    df2 = df.copy()
    for feat in df2.columns:
        for i in range(1, n):
            df2[f"{feat}_{i}"] = df2[feat]
    return df2


def add_cat_feat(df, fname):
    import string

    s = string.ascii_lowercase
    vals = [s[i % 20] for i in range(len(df))]
    df[fname] = vals


# Test Cell
# hide
@pytest.mark.skipif(sys.platform != "linux", reason="plot test only for linux")
def test_return_df_true():
    #   Setup
    #   Run
    corr = plot_feat_correlations(df.head(100), return_df=True)
    #   Verify
    assert isinstance(
        corr, pd.DataFrame
    ), f"should return a dataframe and got {type(corr)}"


@pytest.mark.skipif(sys.platform != "linux", reason="plot test only for linux")
def test_return_df_false():
    #   Setup
    #   Run
    corr = plot_feat_correlations(df.head(100), return_df=False)
    #   Verify
    assert corr is None, f"should return None and got {corr}"


# Test Cell
# hide
@pytest.mark.skipif(sys.platform != "linux", reason="plot test only for linux")
def test_non_numericals_dropped():
    #    Setup
    df_test = df.head(100).copy()
    cat_fnames = [f"categorical_{i}" for i in range(1, 6)]
    for fname in cat_fnames:
        add_cat_feat(df_test, fname)

    #   Run
    corr = plot_feat_correlations(df_test, return_df=True)

    #   Verify
    assert corr.shape[1] == df_test.shape[1] - 5
    assert "categorical" not in " ".join(corr.columns)
